// worker thread code

#include "stdafx.h"
#include "worker.h"


// a note on initialReceiveSize: If the client starts sending
// ass soon as it has a connection, you can leave this value
// != 0. If, however, the client expects the server to offer
// a banner first, you *must* set initialReceiveSize to ZERO.
// If you don't, AcceptEx() will wait for the client to speak
// up, while the client waits for the server to say something!

const int OV::initialReceiveSize = OV::bs - 2 * ( sizeof sockaddr_in + 16 );
const int OV::addrlen = sizeof sockaddr_in + 16;
long runningThreads = 0;
extern SOCKET listener; // the listener socket, obviously.
HANDLE hiocp = 0;


void DoIo( OV &ov );
void DoClose( OV &ov, bool force = false );
void ReinitContext( OV &ov );
void PostRead( OV &ov );
bool DoCommands( OV &ov );
void DoOneCommand( char *p, OV &ov );


unsigned int __stdcall Worker( void *arg )
{
	DWORD n, gle;
	ULONG key;
	BOOL result;
	OV *pov;

	InterlockedIncrement( &runningThreads );
	// attach to IOCP and start looping
	for ( ; ; )
	{
		InterlockedDecrement( &runningThreads );
		// note how I am blithely assuming that a pointer
		// to one of my OV structs is the same as a pointer
		// to the OVERLAPPED class which is OV's parent.
		// Kids, do not do this at home -- _our_ pointers
		// are highly trained athletes.
		result = GetQueuedCompletionStatus( hiocp, &n,
			&key, (OVERLAPPED **) &pov, INFINITE );
		gle = GetLastError();
		if ( pov != 0 )
			pov->n = n;
		InterlockedIncrement( &runningThreads );

		if ( result == FALSE ) // something gone awry?
		{
			// something failed. Was it the call itself, or did we just eat
			// a failed socket I/O?
			if ( pov == 0 ) // the IOCP op failed
			{
				// no key, no byte count, no overlapped info available!
				// how do I handle this?
				// by ignoring it. I don't even have a socket handle to close.
				// You might want to abort instead, as something is seriously
				// wrong if we get here.
			}
			else
			{
				// key, byte count, and overlapped are valid.
				// tear down the connection and requeue the socket
				DoClose( *pov, true );
			}
		}
		else // GQCS() returned TRUE
		{
			// check if we are to die
			if ( key == COMPKEY_DIEDIEDIE )
				break; // sniffle

			// since we are still here, we have an I/O to process
			DoIo( *pov ); // n is redundant -- see pov->InternalHigh
		}
	}

	InterlockedDecrement( &runningThreads );
	cout << "La garde meurt, mais elle ne se rend pas." << endl;

	return 0;
}


void DoClose( OV &ov, bool force /* = false */ )
{
	struct linger li = { 0, 0 }; // default: SO_DONTLINGER

	if ( force )
		li.l_onoff = 1; // SO_LINGER, timeout = 0

	setsockopt( ov.s, SOL_SOCKET, SO_LINGER, (char *) &li, sizeof li );
	closesocket( ov.s );
	if ( ov.sendFile )
		CloseHandle( ov.hf );

	ReinitContext( ov );
}


void ReinitContext( OV &ov )
{
	ov.s = socket( AF_INET, SOCK_STREAM, 0 );
	if ( ov.s != INVALID_SOCKET )
	{
		if ( ! AcceptEx( listener, ov.s, &ov.buf[0],
			ov.bs - ov.addrlen * 2, ov.addrlen,
			ov.addrlen, &ov.n, &ov ) )
		{
			// AcceptEx() returned FALSE, check for ERROR_IO_PENDING
			if ( GetLastError() == ERROR_IO_PENDING )
			{
				ov.state = stAccepting;
				ov.ignore = ov.sendFile = false;
				ov.end = &ov.buf[0];
			}
			else
				clog << "AcceptEx() in ReinitContext() failed, gle == " <<
					GetLastError() << endl;
		}
	}
	else
		clog << "Socket creation in ReinitContext() failed, gle == " <<
			GetLastError() << endl;
}


// handle a successful I/O
void DoIo( OV &ov )
{
	int locallen, remotelen;
	string tt;
	sockaddr_in *plocal = 0, *premote = 0;

	// what was that socket doing, anyway?
reevaluate:
	switch ( ov.state )
	{
	case stAccepting:
		GetAcceptExSockaddrs( &ov.buf[0], ov.bs - ov.addrlen * 2,
			ov.addrlen, ov.addrlen, (sockaddr **) &plocal, &locallen,
			(sockaddr **) &premote, &remotelen );
		memcpy( &ov.local, plocal, sizeof sockaddr_in );
		memcpy( &ov.peer, premote, sizeof sockaddr_in );
		
		// we still need to associate the newly connected socket to outr IOCP:
		CreateIoCompletionPort( (HANDLE) ov.s, hiocp, 0, concurrency );
		tt = inet_ntoa( premote->sin_addr );
		cout << "accepted: from " << tt.c_str() << ":" << ntohs( premote->sin_port ) <<
			" to " << inet_ntoa( plocal->sin_addr ) << ":" << ntohs( plocal->sin_port ) << endl;
		ov.state = stReading;
		if ( ov.n == 0 )
			break; // nothing read yet, don't attempt to buffer it
		goto reevaluate; // process the buffer read by
			// AcceptEx() like a normal read completion

	case stReading:
		if ( ov.n == SOCKET_ERROR ) // error?
		{
			DoClose( ov, true );
		}
		else if ( ov.n == 0 ) // connection closing?
		{
			DoClose( ov, false );
		}
		else
		{
			ov.end += ov.n;
			if ( ! DoCommands( ov ) ) // no commands found? recv() some more!
			{
				// post anoher read, because DoCommands() did not post
				// a write whose completion would cause a read to be posted
				PostRead( ov );
			}
			else
			{
				// (intentionally empty else branch)

				// DoCommands() has posted a write; the completion of
				// this write, handled below, will post another read
			}
		}
		break;

	case stWriting:
		// a write just completed, post another read
		// but check for files to send first
		if ( ov.n == SOCKET_ERROR ) // error?
		{
			DoClose( ov, true );
		}
		else if ( ov.n == 0 ) // connection closing?
		{
			DoClose( ov, false );
		}
		else
		{
			if ( ov.sendFile )
			{
				// ov.hf has a handle to a file to transmit
				memset( &ov.tfb, '\0', sizeof ov.tfb );
				if ( ! ov.tfHead.empty() )
				{
					ov.tfb.HeadLength = ov.tfHead.size();
					ov.tfb.Head = (void *) ov.tfHead.c_str();
				}
				if ( ! ov.tfHead.empty() )
				{
					ov.tfb.TailLength = ov.tfTail.size();
					ov.tfb.Tail = (void *) ov.tfTail.c_str();
				}
				
				if ( TransmitFile( ov.s, ov.hf, 0, 0, &ov, &ov.tfb, 0 ) ||
					GetLastError() == ERROR_IO_PENDING )
				{
					ov.state = stSendingFile;
				}
				else
				{
					// TF() error
					clog << "TransmitFile() failed, gle == " << GetLastError() << endl;
					DoClose( ov, true );
				}
			}
			else
			{
				// the write was OK, post another read
				PostRead( ov );
			}
		}
		break;
	case stSendingFile:
		if ( ov.n == SOCKET_ERROR ) // error?
		{
			DoClose( ov, true );
		}
		else if ( ov.n == 0 ) // connection closing?
		{
			DoClose( ov, false );
		}
		else
		{
			ov.sendFile = false;
			CloseHandle( ov.hf );
			// go back to reading commands
			PostRead( ov );
		}
		break;
	}
}


// returns true if read completed immediately
void PostRead( OV &ov )
{
	// if there is not enough room in the buffer,
	// flush it and set the ignore-this-cmd flag
	if ( ov.end == &ov.buf[ov.bs] )
	{
		ov.end = &ov.buf[0];
		ov.ignore = true;
	}

	ov.state = stReading; // obviously
	ov.n = &ov.buf[ov.bs] - ov.end;
	if ( ReadFile( (HANDLE) ov.s, ov.end, ov.n, &ov.n, &ov ) )
	{
		// everything OK
		return;
	}
	else
	{
		if ( GetLastError() != ERROR_IO_PENDING )
		{
			clog << "PostRead() failed, gle == " << GetLastError() << endl;
			DoClose( ov, true );
		}
		else
		{
			// (else branch intentionally empty)

			// if we get here, gle == ERROR_IO_PENDING,
			// which is fine by me
		}
	}

	return;
}


bool DoCommands( OV &ov )
{
	int numCommands = 0; // number of commands removed from ov.buf
	char *p, *q;

	ov.reply.erase();

	for ( p = &ov.buf[0]; ; )
	{
		// skip line feeds left over from prev command
		for ( ; p < ov.end && *p == '\n'; ++ p )
			;

		// first, locate the end of the command line
		for ( q = p; q < ov.end && *q != '\r'; ++ q )
			;

		if ( q == ov.end ) // no complete command left in the buffer
		{
			if ( p > &ov.buf[0] ) // did we process something already?
			{
				// if so, shift the remaining buffer to the left
				memmove( &ov.buf[0], p, ov.end - p );
				ov.end -= p - &ov.buf[0];
			}
			break;
		}

		// p points at the start of a command, q points at its terminating character
		*q = '\0';

		if ( ov.ignore ) // is this the rest of an overflowing command line?
		{
			ov.ignore = false; // yup, and we did penance
			ov.reply += "One command ignored due to its length.\r\n";
		}
		else // do not ignore this command
		{
			DoOneCommand( p, ov ); // presumably, this adds to ov.reply
		}
		++ numCommands;
		p = q + 1;
	}

	if ( ! ov.reply.empty() )
	{
		ov.state = stWriting;

		// submit a WriteFile() for the reply buffer
		ov.n = ov.reply.size();
		if ( ! WriteFile( (HANDLE) ov.s, ov.reply.c_str(), ov.n, &ov.n, &ov ) )
		{
			if ( GetLastError() != ERROR_IO_PENDING )
			{
				clog << "WriteFile() failed, gle == " << GetLastError() << endl;
				DoClose( ov, true );
				// the fall-through returns true because the caller
				// shouldn't post another read -- the reinitialised
				// socket has an AcceptEx() pending
			}
			else
			{
				// (else branch intentionally empty)

				// if we get here, gle == ERROR_IO_PENDING; nothing
				// left to do but return true. Caller loops back to GQCS().
			}
		}
		else // WriteFile()
		{
			// the write completed immediately
			// doeesn't bother us -- we will still
			// get the completion packet
		}
		return true;
	}
	else // reply was empty
		return false; // tell the caller that we didn't post a write.
			// this means that no write completion will occur, therefore
			// no read will be posted -- the caller needs to do that himself
}


void DoOneCommand( char *cmd, OV &ov )
{
	char *p;
	static const char * const delim = " \t";
	static const string prompt( "\r\n> " ), crlf( "\r\n" );

	ov.reply += prompt + cmd + crlf;
	p = strtok( cmd, delim );

	if ( p == 0 )
		ov.reply += "Nothing to do.\r\n";
	else if ( stricmp( p, "help" ) == 0 || stricmp( p, "?" ) == 0 )
	{
		ov.reply += "Valid commands:\r\n"
			"  help\r\n"
			"  quit\r\n"
			"  sleep <msec>\r\n"
			"  status\r\n"
			"  xmit <filename>\r\n";
	}
	else if ( stricmp( p, "quit" ) == 0 )
	{
		ov.reply += "Death is lighter than a flower;\r\nduty, heavier than a mountain.\r\n";
		SetEvent( suicideEvent );
	}
	else if ( stricmp( p, "status" ) == 0 )
	{
		char tt[40];
		itoa( ntohs( ov.peer.sin_port ), tt, 10 );
		ov.reply += "client address " + string( inet_ntoa( ov.peer.sin_addr ) ) + ":" + tt + crlf;
		itoa( ntohs( ov.local.sin_port ), tt, 10 );
		ov.reply += "server address " + string( inet_ntoa( ov.local.sin_addr ) ) + ":" + tt + crlf;
	}
	else if ( stricmp( p, "sleep" ) == 0 )
	{
		p = strtok( NULL, delim );
		DWORD timeout = ( p == 0 )? 0: atoi( p );
		if ( timeout < 0 || timeout > 60000 )
			timeout = 50;
		Sleep( timeout );
		ov.reply += "Good morning!\r\n";
	}
	else if ( stricmp( p, "xmit" ) == 0 )
	{
		p = strtok( NULL, delim );
		ov.tfHead = string( "And here is your file (" ) + p + string( "):\r\n" ) + 
			"------------------------------------------------------------------------------\r\n";
		ov.tfTail = "------------------------------------------------------------------------------\r\n"
			"There you go!\r\n";
		ov.hf = CreateFile( p, GENERIC_READ, FILE_SHARE_READ, 0,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0 );
		if ( ov.hf == INVALID_HANDLE_VALUE )
		{
			char tt[40];
			itoa( GetLastError(), tt, 10 );
			ov.sendFile = false;
			ov.reply = "Sorry, but the file is unavailable (error " + string( tt ) + ").\r\n"
				"Remember that I am too lazy to correctly handle file names with spaces.\r\n";
		}
		else
		{
			// tell DoIo()'s transmit completion case to start the file transfer
			ov.sendFile = true;
			ov.reply += "Your file is coming up next ...\r\n";
		}
	}
	else
	{
		ov.reply += "Unknown command.\r\n";
	}
}
